//-----------------------------------------------------------------------------
// File: Stretch.CPP
//
// Desc: This program demonstrated clipped blting and stretched clipped blting.
//       It is a non-exclusive mode application that displays a rotating donut
//       in a window.  Clipped blting can be demonstrated by moving another
//       window partially or completely in front of the stretch window.  The
//       rotating donut does not overwrite the clipping window.
//
//       The size of the rotating donut can be changed with menu selections.
//       Any other size than 1x1 demonstrates stretch blting.  The window can
//       also be resized by grabbing any one of the corners with the mouse.
//
//       Another menu option can be used to change the rate of rotation of the
//       donut.
//
//       This is not an exclusive mode application and so it is incapable of
//       setting the display mode.  Therefore, it must be executed on an 8 bit
//       per pixel display.  It will not work correctly with other pixel
//       depths.
//
// Copyright (c) 1995-1999 Microsoft Corporation. All rights reserved.
// Ported by Vadim Bodrov, TMT Development Corp.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Include files
//-----------------------------------------------------------------------------
uses Messages, Windows, DXLite;

{$r stretch.res}

const
    IDI_MAIN_ICON                 = 101;
    IDR_MENU                      = 102;
    ID_FILE_EXIT                  = 10001;
    ID_ROTATION_STOP              = 40001;
    ID_ROTATION_SLOW              = 40011;
    ID_ROTATION_FAST              = 40012;
    ID_SIZE_1X1                   = 40002;
    ID_SIZE_2X1                   = 40003;
    ID_SIZE_3X1                   = 40004;
    ID_SIZE_1X2                   = 40005;
    ID_SIZE_2X2                   = 40006;
    ID_SIZE_3X2                   = 40007;
    ID_SIZE_1X3                   = 40008;
    ID_SIZE_2X3                   = 40009;
    ID_SIZE_3X3                   = 40010;



//-----------------------------------------------------------------------------
// Local definitions
//-----------------------------------------------------------------------------
const
    NAME               = 'Stretch';
    TITLE              = 'Direct Draw Stretch Example';

//-----------------------------------------------------------------------------
// Default settings
//-----------------------------------------------------------------------------
const
    SIZEX              = 64;
    SIZEY              = 64;

//-----------------------------------------------------------------------------
// Global data
//-----------------------------------------------------------------------------
var
    g_pDD: PPDirectDraw7                := nil;     // DirectDraw object
    g_pDDSPrimary: PPDirectDrawSurface7 := nil;     // DirectDraw primary surface
    g_pDDSOne: PPDirectDrawSurface7     := nil;     // Offscreen surface 1
    g_pClipper: PPDirectDrawClipper     := nil;     // Clipper for primary
    g_pDDPal: PPDirectDrawPalette       := nil;     // The primary surface palette
    g_bActive: Boolean                  := FALSE;   // Is application active?



//-----------------------------------------------------------------------------
// Local data
//-----------------------------------------------------------------------------
// Name of our bitmap resource.
var
    szBitmap: PChar := 'DONUT';
    dwUpdateDelay: DWORD := 13;




//-----------------------------------------------------------------------------
// Name: ReleaseAllObjects()
// Desc: Finished with all objects we use; release them
//-----------------------------------------------------------------------------
procedure ReleaseAllObjects;
begin
    if (g_pDD <> nil) then
    begin
        if (g_pDDSPrimary <> nil) then
        begin
            g_pDDSPrimary^^.Release^(g_pDDSPrimary);
            g_pDDSPrimary := nil;
        end;
        if (g_pDDSOne <> nil) then
        begin
            g_pDDSOne^^.Release^(g_pDDSOne);
            g_pDDSOne := nil;
        end;
        if (g_pDDPal <> nil) then
        begin
            g_pDDPal^^.Release^(g_pDDPal);
            g_pDDPal := nil;
        end;
        g_pDD^^.Release^(g_pDD);
        g_pDD := nil;
    end;
end;




//-----------------------------------------------------------------------------
// Name: InitFail()
// Desc: This function is called if an initialization function fails
//-----------------------------------------------------------------------------
function InitFail(Wnd: HWND; hRet: HRESULT; szError: LPCTSTR): HResult;
begin
    ReleaseAllObjects;
    MessageBox(Wnd, szError, TITLE, MB_OK);
    DestroyWindow(Wnd);
    Result := hRet;
end;




//-----------------------------------------------------------------------------
// Name: RestoreAll()
// Desc: Restore all lost objects
//-----------------------------------------------------------------------------
function RestoreAll: Boolean;
begin
    Result := (g_pDDSPrimary^^.Restore^(g_pDDSPrimary) = DD_OK) and
              (g_pDDSOne^^.Restore^(g_pDDSOne) = DD_OK) and
              (DDReLoadBitmap(g_pDDSOne, szBitmap) = DD_OK);
end;




//-----------------------------------------------------------------------------
// Name: UpdateFrame()
// Desc: Decide what needs to be blitted next, wait for flip to complete,
//       then flip the buffers.
//-----------------------------------------------------------------------------
var
    lastTickCount:  DWORD := 0;
    currentFrame:   Longint := 0;
    haveBackground: Boolean := FALSE;

procedure UpdateFrame(Wnd: HWND);
var
    thisTickCount: DWORD;
    rcRect: TRect;
    destRect: TRect;
    hRet: HResult;
    pt: TPoint;
begin
    thisTickCount := GetTickCount;
    if ((thisTickCount - lastTickCount) <= dwUpdateDelay)  then
        exit;

    // Move to next frame;
    lastTickCount := thisTickCount;
    currentFrame +:= 1;
    if (currentFrame > 59) then
        currentFrame := 0;

    // Blit the stuff for the next frame
    rcRect.left := currentFrame mod 10 * 64;
    rcRect.top := currentFrame div 10 * 64;
    rcRect.right := currentFrame mod 10 * 64 + 64;
    rcRect.bottom := currentFrame div 10 * 64 + 64;

    GetClientRect(Wnd, destRect);
    if (destRect.right < 128) then
        destRect.right := 64;
    if (destRect.bottom < 64) then
        destRect.bottom := 64;

    pt.x := 0; pt.y := 0;
    ClientToScreen(Wnd, pt);
    OffsetRect(destRect, pt.x, pt.y);

    while (TRUE) do
    begin
        hRet := g_pDDSPrimary^^.Blt^(g_pDDSPrimary, @destRect, g_pDDSOne, @rcRect, 0, nil);
        if (hRet = DD_OK) then
            break;

        if (hRet = DDERR_SURFACELOST) then
        begin
            if (not RestoreAll) then
                exit;
        end;
        if (hRet <> DDERR_WASSTILLDRAWING) then
            exit;
    end;
    if (hRet <> DD_OK) then
        exit;
end;




//-----------------------------------------------------------------------------
// Name: WindowProc()
// Desc: The Main Window Procedure
//-----------------------------------------------------------------------------
function WindowProc conv arg_stdcall (Wnd: HWND; message: UINT; wParam: WPARAM; lParam: LPARAM): LResult;
var
    rc:      TRect;
    xRight:  Longint := -1;
    yBottom: Longint := -1;
    pMinMax: PMinMaxInfo;
begin
    case message of
        WM_ACTIVATE:
        begin
            // Pause if minimized
            g_bActive := not Boolean(HIWORD(wParam));
            Result := 0;
            exit;
        end;

        WM_COMMAND:
        begin
            // Handle all menu and accelerator commands
            case (LOWORD(wParam)) of
                ID_ROTATION_STOP:
                    dwUpdateDelay := $7fffffff;
                ID_ROTATION_SLOW:
                    dwUpdateDelay := 200;
                ID_ROTATION_FAST:
                    dwUpdateDelay := 13;
                ID_FILE_EXIT:
                begin
                    PostMessage(Wnd, WM_CLOSE, 0, 0);
                    Result := 0;
                    exit;
                end;
                ID_SIZE_1X1:
                begin
                    xRight := SIZEX * 1;
                    yBottom := SIZEY * 1;
                end;
                ID_SIZE_2X1:
                begin
                    xRight := SIZEX * 2;
                    yBottom := SIZEY * 1;
                end;
                ID_SIZE_3X1:
                begin
                    xRight := SIZEX * 3;
                    yBottom := SIZEY * 1;
                end;
                ID_SIZE_1X2:
                begin
                    xRight := SIZEX * 1;
                    yBottom := SIZEY * 2;
                end;
                ID_SIZE_2X2:
                begin
                    xRight := SIZEX * 2;
                    yBottom := SIZEY * 2;
                end;
                ID_SIZE_3X2:
                begin
                    xRight := SIZEX * 3;
                    yBottom := SIZEY * 2;
                end;
                ID_SIZE_1X3:
                begin
                    xRight := SIZEX * 1;
                    yBottom := SIZEY * 3;
                end;
                ID_SIZE_2X3:
                begin
                    xRight := SIZEX * 2;
                    yBottom := SIZEY * 3;
                end;
                ID_SIZE_3X3:
                begin
                    xRight := SIZEX * 3;
                    yBottom := SIZEY * 3;
                end;
            end;
            if (xRight <> -1) then
            begin
                // Change the window size if set
                SetRect(rc, 0, 0, xRight, yBottom);
                AdjustWindowRectEx(rc,
                                   GetWindowLong(Wnd, GWL_STYLE),
                                   GetMenu(Wnd) <> 0,
                                   GetWindowLong(Wnd, GWL_EXSTYLE));
                SetWindowPos(Wnd, 0, 0, 0, rc.right - rc.left,
                             rc.bottom - rc.top,
                             SWP_NOZORDER or SWP_NOMOVE or SWP_NOACTIVATE);
                Result := 0;
                exit;
            end;
        end;

        WM_DESTROY:
        begin
            // Clean up and close the app
            ReleaseAllObjects;
            PostQuitMessage(0);
            Result := 0;
            exit;
        end;

        WM_GETMINMAXINFO:
        begin
            // Fix the minimum size of the window to SIZEX x SIZEY
            pMinMax := PMinMaxInfo(lParam);
            pMinMax^.ptMinTrackSize.x := SIZEX + GetSystemMetrics(SM_CXSIZEFRAME) * 2;
            pMinMax^.ptMinTrackSize.y := SIZEY +
                                        GetSystemMetrics(SM_CYSIZEFRAME) * 2 +
                                        GetSystemMetrics(SM_CYCAPTION) +
                                        GetSystemMetrics(SM_CYMENU);
            Result :=0;
            exit;
        end;

        WM_KEYDOWN:
        begin
            // Handle any non-accelerated key commands
            if wParam in [VK_ESCAPE, VK_F12]  then
            begin
                PostMessage(Wnd, WM_CLOSE, 0, 0);
                Result := 0;
                exit;
            end;
        end;

        WM_PALETTECHANGED:
            if (wParam = Wnd) then
            begin
               // Fall through to WM_QUERYNEWPALETTE
            end;
        WM_QUERYNEWPALETTE:
        begin
            // Install our palette here
            if (g_pDDPal <> nil) then
            begin
                g_pDDSPrimary^^.SetPalette^(g_pDDSPrimary, g_pDDPal);
                // Reload the bitmap into the surface because the palette
                // has changed..
                DDReLoadBitmap(g_pDDSOne, szBitmap);
                Result := 0;
            end;
        end;
    end;
    Result := DefWindowProc(Wnd, message, wParam, lParam);
end;




//-----------------------------------------------------------------------------
// Name: InitApp()
// Desc: Do work required for every instance of the application:
//          Create the window, initialize data
//-----------------------------------------------------------------------------
function InitApp(var phWnd: HWND): HResult;
var
    wc:   TWndClass;
    ddsd: TDDSurfaceDesc2;
    hRet: HResult;
begin
    // Set up and register window class
    wc.style := CS_HREDRAW or CS_VREDRAW;
    wc.lpfnWndProc := @WindowProc;
    wc.cbClsExtra := 0;
    wc.cbWndExtra := 0;
    wc.hInstance := hInstance;
    wc.hIcon := LoadIcon(hInstance, MAKEINTRESOURCE(IDI_MAIN_ICON));
    wc.hCursor := LoadCursor(0, IDC_ARROW);
    wc.hbrBackground := (COLOR_WINDOW + 1);
    wc.lpszMenuName := MAKEINTRESOURCE(IDR_MENU);
    wc.lpszClassName := NAME;
    RegisterClass(wc);

    // Create a window
    phWnd := CreateWindowEx(0,
                            NAME,
                            TITLE,
                            WS_OVERLAPPEDWINDOW,
                            CW_USEDEFAULT,
                            CW_USEDEFAULT,
                            128,
                            128,
                            0,
                            0,
                            hInstance,
                            nil);
    if (phWnd = 0) then
    begin
        Result := DDERR_GENERIC;
        exit;
    end;
    PostMessage(phWnd, WM_COMMAND, ID_SIZE_1X1, 0);
    ShowWindow(phWnd, CmdShow);
    UpdateWindow(phWnd);

    ///////////////////////////////////////////////////////////////////////////
    // Create the main DirectDraw object
    ///////////////////////////////////////////////////////////////////////////
    hRet := DirectDrawCreateEx(nil, g_pDD, IID_IDirectDraw7, nil);
    if (hRet <> DD_OK) then
    begin
        Result := InitFail(phWnd, hRet, 'DirectDrawCreateEx FAILED');
        exit;
    end;

    // Get normal mode
    hRet := g_pDD^^.SetCooperativeLevel^(g_pDD, phWnd, DDSCL_NORMAL);
    if (hRet <> DD_OK) then
    begin
        Result := InitFail(phWnd, hRet, 'SetCooperativeLevel FAILED');
        exit;
    end;

    // Create the primary surface
    ZeroMemory(@ddsd, sizeof(ddsd));
    ddsd.dwSize := sizeof(ddsd);
    ddsd.dwFlags := DDSD_CAPS;
    ddsd.ddsCaps.dwCaps := DDSCAPS_PRIMARYSURFACE;
    hRet := g_pDD^^.CreateSurface^(g_pDD, ddsd, g_pDDSPrimary, nil);
    if (hRet <> DD_OK) then
    begin
        Result := InitFail(phWnd, hRet, 'CreateSurface FAILED');
        exit;
    end;

    // Create a clipper for the primary surface
    hRet := g_pDD^^.CreateClipper^(g_pDD, 0, g_pClipper, nil);
    if (hRet <> DD_OK) then
    begin
        Result := InitFail(phWnd, hRet, 'CreateClipper FAILED');
        exit;
    end;
    hRet := g_pClipper^^.SetHWnd^(g_pClipper, 0, phWnd);
    if (hRet <> DD_OK) then
    begin
        Result := InitFail(phWnd, hRet, 'SetHWnd FAILED');
        exit;
    end;
    hRet := g_pDDSPrimary^^.SetClipper^(g_pDDSPrimary, g_pClipper);
    if (hRet <> DD_OK) then
    begin
        Result := InitFail(phWnd, hRet, 'SetClipper FAILED');
        exit;
    end;

    // Create and set the palette
    g_pDDPal := DDLoadPalette(g_pDD, szBitmap);
    if (g_pDDPal <> nil) then
        g_pDDSPrimary^^.SetPalette^(g_pDDSPrimary, g_pDDPal);

    // Load our bitmap
    g_pDDSOne := DDLoadBitmap(g_pDD, szBitmap, 0, 0);
    if (g_pDDSOne = nil) then
    begin
        Result := InitFail(phWnd, hRet, 'DDLoadBitmap FAILED');
        exit;
    end;

    Result := DD_OK;
end;




//-----------------------------------------------------------------------------
// Name: WinMain()
// Desc: Initialization, message loop
//-----------------------------------------------------------------------------
var
    msg: TMsg;
    Wnd: HWND;
begin
    if (InitApp(Wnd) <> DD_OK) then exit;

    while (TRUE) do
    begin
        if (PeekMessage(msg, 0, 0, 0, PM_NOREMOVE)) then
        begin
            if (not GetMessage(msg, 0, 0, 0)) then
                exit;
            TranslateMessage(msg);
            DispatchMessage(msg);
        end else
        if (g_bActive) then
        begin
            UpdateFrame(Wnd);
        end else
        begin
            // Make sure we go to sleep if we have nothing else to do
            WaitMessage;
        end;
    end;
end.
